/*
* Copyright (c) 2020 Oracle and/or its affiliates.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

def aetherScript = '''
/*
* Copyright (c) 2020 Oracle and/or its affiliates.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

import org.apache.maven.repository.internal.MavenRepositorySystemUtils
import org.apache.maven.wagon.Wagon
import org.apache.maven.wagon.providers.http.HttpWagon
import org.codehaus.plexus.DefaultPlexusContainer
import org.codehaus.plexus.component.repository.ComponentDescriptor
import org.eclipse.aether.RepositorySystem
import org.eclipse.aether.RepositorySystemSession
import org.eclipse.aether.artifact.DefaultArtifact
import org.eclipse.aether.collection.CollectRequest
import org.eclipse.aether.connector.basic.BasicRepositoryConnectorFactory
import org.eclipse.aether.graph.Dependency
import org.eclipse.aether.internal.transport.wagon.PlexusWagonProvider
import org.eclipse.aether.repository.LocalRepository
import org.eclipse.aether.repository.Proxy
import org.eclipse.aether.repository.RemoteRepository
import org.eclipse.aether.repository.RepositoryPolicy
import org.eclipse.aether.resolution.ArtifactRequest
import org.eclipse.aether.resolution.DependencyRequest
import org.eclipse.aether.spi.connector.RepositoryConnectorFactory
import org.eclipse.aether.spi.connector.transport.TransporterFactory
import org.eclipse.aether.transport.wagon.WagonProvider
import org.eclipse.aether.transport.wagon.WagonTransporterFactory
import org.eclipse.aether.util.filter.DependencyFilterUtils
import org.eclipse.aether.util.repository.AuthenticationBuilder

/**
 * Standalone aether utility.
 */
class AetherImpl {

    private List<RemoteRepository> remoteRepos
    private RepositorySystem repoSystem
    private RepositorySystemSession repoSession

    @SuppressWarnings("GroovyAssignabilityCheck")
    AetherImpl(File localRepoDir, remoteArtifactRepos) {
        def plexusContainer = new DefaultPlexusContainer()
        plexusContainer.addComponentDescriptor(plexusDesc(Wagon.class, "http", HttpWagon.class, "per-lookup"))
        plexusContainer.addComponentDescriptor(plexusDesc(Wagon.class, "https", HttpWagon.class, "per-lookup"))
        repoSystem = MavenRepositorySystemUtils.newServiceLocator()
                .setServices(WagonProvider.class, new PlexusWagonProvider(plexusContainer))
                .addService(TransporterFactory.class, WagonTransporterFactory.class)
                .addService(RepositoryConnectorFactory.class, BasicRepositoryConnectorFactory.class)
                .getService(RepositorySystem.class)
        repoSession = MavenRepositorySystemUtils.newSession()
        repoSession.setLocalRepositoryManager(repoSystem.newLocalRepositoryManager(repoSession,
                new LocalRepository(localRepoDir)))
        remoteRepos = remoteArtifactRepos.collect { remoteRepo(it) }
    }

    private static ComponentDescriptor plexusDesc(Class<?>roleClass,
                                                  String roleHint,
                                                  Class<?> implClass,
                                                  String strategy) {
        def desc = new ComponentDescriptor()
        desc.setRoleClass(roleClass)
        desc.setRoleHint(roleHint)
        desc.setImplementationClass(implClass)
        desc.setInstantiationStrategy(strategy)
        return desc
    }

    @SuppressWarnings("GroovyAssignabilityCheck")
    private static RemoteRepository remoteRepo(aRepo) {
        def builder = new RemoteRepository.Builder(aRepo.getId(), aRepo.getLayout().getId(), aRepo.getUrl());
        def releases = aRepo.getReleases()
        if (releases != null) {
            def releasePolicy = new RepositoryPolicy(releases.isEnabled(), releases.getUpdatePolicy(),
                    releases.getChecksumPolicy())
            builder.setReleasePolicy(releasePolicy)
        }
        def snapshots = aRepo.getSnapshots()
        if (snapshots != null) {
            def snapshotPolicy = new RepositoryPolicy(snapshots.isEnabled(), snapshots.getUpdatePolicy(),
                    snapshots.getChecksumPolicy())
            builder.setSnapshotPolicy(snapshotPolicy)
        }
        def proxy = aRepo.getProxy()
        if (proxy != null) {
            def authentication = new AuthenticationBuilder()
                    .addUsername(proxy.getUserName())
                    .addPassword((String) proxy.getPassword())
                    .build()
            builder.setProxy(new Proxy(proxy.getProtocol(), proxy.getHost(), proxy.getPort(), authentication));
        }
        return builder.build();
    }

    /**
     * Resolve the given artifact.
     *
     * @param groupId the groupId
     * @param artifactId the artifactId
     * @param type the type e.g. {@code "jar"}
     * @param version the version
     * @return artifact file
     */
    File resolveArtifact(String groupId, String artifactId, String type, String version) {
        return repoSystem
                .resolveArtifact(repoSession, new ArtifactRequest()
                        .setArtifact(new DefaultArtifact(groupId, artifactId, type, version)))
                .getArtifact()
                .getFile()
    }

    /**
     * Resolve transitive dependencies of the given GAV.
     *
     * @param gav GAV string (groupId:artifactId:version)
     * @return list of files
     */
    List<File> resolveDependencies(String gav) {
        return repoSystem
                .resolveDependencies(repoSession, new DependencyRequest()
                        .setCollectRequest(new CollectRequest()
                                .setRoot(new Dependency(new DefaultArtifact(gav), "compile"))
                                .setRepositories(remoteRepos))
                        .setFilter(DependencyFilterUtils.classpathFilter("runtime")))
                .getArtifactResults()
                .collect { it.getArtifact().getFile() }
    }
}

/**
 * Create a new Aether instance.
 * @param localRepoDir local repository directory
 * @param remoteArtifactRepos remote artifacts repositories
 * @return AetherImpl
 */
def create(localRepoDir, remoteArtifactRepos) {
    return new AetherImpl(localRepoDir, remoteArtifactRepos)
}

return this
'''

def engineScript = '''
/*
* Copyright (c) 2020 Oracle and/or its affiliates.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

import org.apache.maven.artifact.versioning.ComparableVersion

class HelidonEngineImpl {

    /**
     * Class loader with the maven core libraries found under {@code ${maven.home}/lib}.
     * It is used to by-pass the restricted plugin class-loader that the {@code archetype-post-generate.groovy} script
     * is executed with.
     */
    private final URLClassLoader mavenCl

    HelidonEngineImpl() {
        List<URL> mavenLibs = new File(System.getProperty("maven.home"), "lib").listFiles()
                .collect{ it.toURI().toURL() }
        mavenCl = new URLClassLoader(mavenLibs.toArray(new URL[mavenLibs.size()]), this.getClass().getClassLoader())
    }

    /**
     * Get the Maven version from the maven libs class loader.
     * The implementation follows the same logic as {@code org.apache.maven.rtinfo.internal.DefaultRuntimeInformation}.
     * @return version string, empty if not found or invalid
     */
    private String getMavenVersion() {
        Properties props = new Properties()
        props.load(mavenCl.getResourceAsStream("META-INF/maven/org.apache.maven/maven-core/pom.properties"))
        String version = props.getProperty( "version", "").trim()
        return version.startsWith('${') ? "" : version
    }

    /**
     * Check that the Maven version is greater or equal to 3.2.5.
     */
    void checkMavenVersion() {
        String mavenVersion = getMavenVersion()
        ComparableVersion minMavenVersion = new ComparableVersion("3.2.5")
        if (!mavenVersion.isEmpty() && new ComparableVersion(mavenVersion) < minMavenVersion) {
            throw new IllegalStateException("Requires Maven >= 3.2.5")
        }
    }

    /**
     * Check that the Java version is greater or equal to 11.
     */
    void checkJavaVersion() {
        try {
            if (Runtime.class.getMethod("version") != null
                    && Runtime.Version.getMethod("feature") != null
                    && Runtime.getRuntime().version().feature() < 11) {
                throw new IllegalStateException()
            }
        } catch (NoSuchMethodException | IllegalStateException ex) {
            throw new IllegalStateException("Requires Java >= 11")
        } catch (Throwable ex) {
            throw new IllegalStateException("Unable to verify Java version", ex)
        }
    }

    /**
     * Generate a project.
     *
     * @param aetherScript script for {@code Aether.groovy}
     * @param localRepo local repository directory
     * @param remoteRepos remote artifacts repositories
     * @param archetypeGroupId archetype groupId
     * @param archetypeArtifactId archetype artifactId
     * @param archetypeVersion archetype version
     * @param engineGav helidon engine GAV string
     * @param props properties
     * @param projectDir project directory
     */
    void generate(String aetherScript,
                  File localRepo,
                  remoteRepos,
                  String archetypeGroupId,
                  String archetypeArtifactId,
                  String archetypeVersion,
                  String engineGav,
                  Map<String, String> props,
                  File projectDir) {

        // set the context class loader for plexus to work properly.
        Thread.currentThread().setContextClassLoader(mavenCl)

        def aether = new GroovyShell(mavenCl)
                .parse(aetherScript)
                .create(localRepo, remoteRepos)

        // resolve the archetype JAR from the local repository
        def archetypeFile = aether
                .resolveArtifact(archetypeGroupId, archetypeArtifactId, "jar", archetypeVersion)

        // resolve the helidon engine libs from remote repository
        List<URL> engineLibs = aether
                .resolveDependencies(engineGav)
                .collect { it.toURI().toURL() }

        // create a class-loader with the engine dependencies
        def ecl = new URLClassLoader(engineLibs.toArray(new URL[engineLibs.size()]), this.getClass().getClassLoader())

        // instantiate the engine
        def engine = ecl.loadClass("io.helidon.build.archetype.engine.ArchetypeEngine")
                .getConstructor(File.class, Map.class)
                .newInstance(archetypeFile, props)

        // delete place place-holder pom
        new File(projectDir, "pom.xml")
                .delete()

        engine.generate(projectDir)
    }
}

/**
 * Create a new engine instance.
 * @return HelidonEngineImpl
 */
def create() {
    return new HelidonEngineImpl()
}

return this
'''

def helidonEngine = new GroovyShell().parse(engineScript).create()
helidonEngine.checkMavenVersion()
helidonEngine.checkJavaVersion()

def engineGav = "io.helidon.build-tools.archetype:helidon-archetype-engine:2.1.2"

def archetypeProperties = request.getProperties()
def props = [
"maven": "true",
    "groupId": archetypeProperties.getProperty("groupId"),
    "artifactId": archetypeProperties.getProperty("artifactId"),
    "version": archetypeProperties.getProperty("version"),
    "name": archetypeProperties.getProperty("name"),
    "package": archetypeProperties.getProperty("package"),
]

def localRepo = request
    .getProjectBuildingRequest()
    .getRepositorySession()
    .getLocalRepository()
    .getBasedir()

def remoteRepos = request
    .getProjectBuildingRequest()
    .getRemoteRepositories()

def projectDir = new File(request.getOutputDirectory() + "/" + request.getArtifactId())

helidonEngine.generate(
    aetherScript,
    localRepo,
    remoteRepos,
    request.getArchetypeGroupId(),
    request.getArchetypeArtifactId(),
    request.getArchetypeVersion(),
    engineGav,
    props,
    projectDir)

